/* 
 * delta_stream.c -- Internal text delta representation
 * 
 * ================================================================
 * Copyright (c) 2000 Collab.Net.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 * any, must include the following acknowlegement: "This product includes
 * software developed by Collab.Net (http://www.Collab.Net/)."
 * Alternately, this acknowlegement may appear in the software itself, if
 * and wherever such third-party acknowlegements normally appear.
 * 
 * 4. The hosted project names must not be used to endorse or promote
 * products derived from this software without prior written
 * permission. For written permission, please contact info@collab.net.
 * 
 * 5. Products derived from this software may not use the "Tigris" name
 * nor may "Tigris" appear in their names without prior written
 * permission of Collab.Net.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL COLLAB.NET OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ====================================================================
 * 
 * This software may consist of voluntary contributions made by many
 * individuals on behalf of Collab.Net.
 */



#include <stdlib.h>
#include "delta_stream.h"
#include "svn_error.h"
#include "svn_string.h"



svn_error_t*
svn_tdelta__init_winfo (svn_delta_stream_t* stream,
                        svn_tdelta__winfo_t* winfo)
{
  apr_pool_t* const pool = apr_make_sub_pool (stream->pool, NULL);
  if (!pool)
    return svn_create_error (APR_ENOMEM, 0,
                             "creating pool for delta window",
                             NULL, stream->pool);
  winfo->window = apr_palloc (pool, sizeof (*winfo->window));
  if (!winfo->window)
    return svn_create_error (APR_ENOMEM, 0,
                             "allocating delta window",
                             NULL, pool);

  winfo->window->num_ops = 0;
  winfo->window->ops = NULL;
  winfo->window->new = svn_string_create ("", pool);
  winfo->window->pool = pool;
  winfo->ops_space = 0;
  winfo->data_offset = 0;
  return NULL;
}



svn_error_t*
svn_tdelta__insert_op (svn_tdelta__winfo_t* winfo,
                       int opcode,
                       apr_off_t offset,
                       apr_off_t length,
                       char const* new_data)
{
  svn_delta_op_t* op;

  /* Create space for the new op. */
  if (winfo->window->num_ops == winfo->ops_space)
    {
      svn_delta_op_t* const old_ops = winfo->window->ops;
      int const new_ops_space = (winfo->ops_space == 0
                                ? 16 : 2 * winfo->ops_space);
      winfo->window->ops =
        apr_palloc (winfo->window->pool,
                    new_ops_space * sizeof (*winfo->window->ops));
      if (!winfo->window->ops)
        return svn_create_error (APR_ENOMEM, 0,
                                 "allocating space for ops in delta windww",
                                 NULL, winfo->window->pool);

      /* Copy any existing ops into the new array */
      if (old_ops)
        memcpy (winfo->window->ops, old_ops,
                winfo->ops_space * sizeof (*winfo->window->ops));

      winfo->ops_space = new_ops_space;
    }

  /* Insert the op. svn_delta_source and svn_delta_target are
     just inserted. For svn_delta_new, the new data must be
     copied into the window. */
  op = &winfo->window->ops[winfo->window->num_ops];
  switch (opcode)
    {
    case svn_delta_source:
    case svn_delta_target:
      op->op = opcode;
      op->offset = offset;
      op->length = length;
      break;
    case svn_delta_new:
      op->op = opcode;
      op->offset = winfo->data_offset;
      op->length = length;
      svn_string_appendbytes (winfo->window->new, new_data,
                              length, winfo->window->pool);
      winfo->data_offset += length;
      break;
    default:
      /* FIXME: Should go barmy here. */
    }

  ++winfo->window->num_ops;
  return NULL;
}



svn_error_t*
svn_text_delta (svn_delta_read_fn_t* source_fn,
                void* source_data,
                svn_delta_read_fn_t* target_fn,
                void* target_data,
                svn_delta_stream_t** stream)
{
  apr_pool_t* const pool = apr_make_sub_pool (NULL, NULL);
  if (!pool)
    return svn_create_error (APR_ENOMEM, 0,
                             "creating pool for delta stream",
                             NULL, NULL);
  
  *stream = apr_palloc (pool, sizeof (**stream));
  if (!*stream)
    return svn_create_error (APR_ENOMEM, 0,
                             "allocating delta stream",
                             NULL, pool);

  (*stream)->source_fn = source_fn; 
  (*stream)->source_data = source_data;
  (*stream)->target_fn = target_fn;
  (*stream)->target_data = target_data;
  (*stream)->pool = pool;
  (*stream)->more = TRUE;
  return NULL;
}


void
svn_free_delta_stream (svn_delta_stream_t* stream)
{
  if (stream)
    apr_destroy_pool (stream->pool);
}


svn_error_t*
svn_next_delta_window (svn_delta_stream_t* stream,
                       svn_delta_window_t** window)
{
  if (!stream->more)
    {
      *window = NULL;
      return NULL;
    }
  else
    {
      svn_tdelta__winfo_t winfo;
      svn_error_t* err = NULL;

      char buffer[57];
      apr_off_t len;

      err = svn_tdelta__init_winfo (stream, &winfo);
      if (err)
        return err;

      *window = winfo.window;
      stream->more = FALSE;

      for (len = sizeof (buffer); ; len = sizeof (buffer))
        {
          stream->target_fn (stream->target_data, buffer, &len);
          if (len == 0)
            break;
          err = svn_tdelta__insert_op (&winfo, svn_delta_new, 0, len, buffer);
          if (err)
            return err;
        }

      return NULL;
    }
}


void
svn_free_delta_window (svn_delta_window_t* window)
{
  if (window)
    apr_destroy_pool (window->pool);
}



/* 
 * local variables:
 * eval: (load-file "../svn-dev.el")
 * end:
 */
